! file src_equations/equations_module.f90
!
! Copyright 2009,2010 Dalton Harvie (daltonh@unimelb.edu.au)
! 
! This file is part of arb finite volume solver, referred to as `arb'.
! 
! arb is a software package designed to solve arbitrary partial
! differential equations on unstructured meshes using the finite volume
! method.  Primarily it consists of fortran source code, perl source
! code and shell scripts.  arb replies on certain third party software
! to run, most notably the computer algebra system maxima
! <http://maxima.sourceforge.net/> which is released under the GNU GPL.
! 
! The copyright of arb is held by Dalton Harvie.
! 
! arb is released under the GNU GPL.  arb is free software: you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License (version 3) as published by the Free Software Foundation.
! You should have received a copy of the GNU General Public License
! along with arb (see file license/gpl.txt after unpacking).  If not,
! see <http://www.gnu.org/licenses/>.
! 
! arb is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
! for more details.
! 
! For full details of arb's license see the license directory.
! 
! The current homepage for the arb finite volume solver project is
! <http://www.chemeng.unimelb.edu.au/people/staff/daltonh/downloads/arb>.
!
!-------------------------------------------------------------------------
module equations_module

contains

!----------------------------------------------------------------------------

subroutine allocate_meta_arrays

! here we allocate array elements of the meta data that corresponds to the input files

use general_module
logical :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine allocate_meta_arrays'

! allocate meta arrays

! allocate var and compound arrays
allocate(var(7))
allocate(compound(7))

!constant: <D> -> var[1] -> var(1) -> var(1)%funk(1)%v
var(1)%name='<D>'
var(1)%multiplier=1.d0
var(1)%units='m^2/s'
var(1)%centring='none'
var(1)%type='constant'
var(1)%rank='scalar'
var(1)%region=''
var(1)%compound_name='<D>'
allocate(var(1)%component(1))
var(1)%component=[ 1 ]
var(1)%output=.false.

!constant: <hole flux> -> var[2] -> var(2) -> var(2)%funk(1)%v
var(2)%name='<hole flux>'
var(2)%multiplier=1.d0
var(2)%units='W/m^2'
var(2)%centring='none'
var(2)%type='constant'
var(2)%rank='scalar'
var(2)%region=''
var(2)%compound_name='<hole flux>'
allocate(var(2)%component(1))
var(2)%component=[ 2 ]
var(2)%output=.false.

!independent: <T> -> var[i,3] -> var(i,3) -> var(3)%funk(nsvar(3,i))%v
var(3)%name='<T>'
var(3)%multiplier=1.d0
var(3)%units='K'
var(3)%centring='cell'
var(3)%type='independent'
var(3)%rank='scalar'
var(3)%region='<all cells>'
var(3)%compound_name='<T>'
allocate(var(3)%component(1))
var(3)%component=[ 3 ]
var(3)%output=.false.
var(3)%magnitude=1.d+2

!equation: <T transport> -> var[i,4] -> var(i,4) -> var(4)%funk(nsvar(4,i))%v
var(4)%name='<T transport>'
var(4)%multiplier=1.d0
var(4)%units='1'
var(4)%centring='cell'
var(4)%type='equation'
var(4)%rank='scalar'
var(4)%region='<domain>'
var(4)%compound_name='<T transport>'
allocate(var(4)%component(1))
var(4)%component=[ 4 ]
var(4)%output=.false.

!equation: <T hole> -> var[j,5] -> var(j,5) -> var(5)%funk(nsvar(5,j))%v
var(5)%name='<T hole>'
var(5)%multiplier=1.d0
var(5)%units='1'
var(5)%centring='face'
var(5)%type='equation'
var(5)%rank='scalar'
var(5)%region='<hole>'
var(5)%compound_name='<T hole>'
allocate(var(5)%component(1))
var(5)%component=[ 5 ]
var(5)%output=.false.

!equation: <T walls> -> var[j,6] -> var(j,6) -> var(6)%funk(nsvar(6,j))%v
var(6)%name='<T walls>'
var(6)%multiplier=1.d0
var(6)%units='1'
var(6)%centring='face'
var(6)%type='equation'
var(6)%rank='scalar'
var(6)%region='<walls>'
var(6)%compound_name='<T walls>'
allocate(var(6)%component(1))
var(6)%component=[ 6 ]
var(6)%output=.false.

!output: <T average> -> var[7] -> var(7) -> var(7)%funk(1)%v
var(7)%name='<T average>'
var(7)%multiplier=1.d0
var(7)%units='1'
var(7)%centring='none'
var(7)%type='output'
var(7)%rank='scalar'
var(7)%region=''
var(7)%compound_name='<T average>'
allocate(var(7)%component(1))
var(7)%component=[ 7 ]
var(7)%output=.false.

!compound: <D> of scalar rank
compound(1)%name='<D>'
compound(1)%units='m^2/s'
compound(1)%centring='none'
compound(1)%type='constant'
compound(1)%rank='scalar'
compound(1)%region=''
allocate(compound(1)%component(1))
compound(1)%component=[ 1 ]
compound(1)%output=.false.

!compound: <hole flux> of scalar rank
compound(2)%name='<hole flux>'
compound(2)%units='W/m^2'
compound(2)%centring='none'
compound(2)%type='constant'
compound(2)%rank='scalar'
compound(2)%region=''
allocate(compound(2)%component(1))
compound(2)%component=[ 2 ]
compound(2)%output=.false.

!compound: <T> of scalar rank
compound(3)%name='<T>'
compound(3)%units='K'
compound(3)%centring='cell'
compound(3)%type='independent'
compound(3)%rank='scalar'
compound(3)%region='<all cells>'
allocate(compound(3)%component(1))
compound(3)%component=[ 3 ]
compound(3)%output=.true.

!compound: <T transport> of scalar rank
compound(4)%name='<T transport>'
compound(4)%units='1'
compound(4)%centring='cell'
compound(4)%type='equation'
compound(4)%rank='scalar'
compound(4)%region='<domain>'
allocate(compound(4)%component(1))
compound(4)%component=[ 4 ]
compound(4)%output=.false.

!compound: <T hole> of scalar rank
compound(5)%name='<T hole>'
compound(5)%units='1'
compound(5)%centring='face'
compound(5)%type='equation'
compound(5)%rank='scalar'
compound(5)%region='<hole>'
allocate(compound(5)%component(1))
compound(5)%component=[ 5 ]
compound(5)%output=.false.

!compound: <T walls> of scalar rank
compound(6)%name='<T walls>'
compound(6)%units='1'
compound(6)%centring='face'
compound(6)%type='equation'
compound(6)%rank='scalar'
compound(6)%region='<walls>'
allocate(compound(6)%component(1))
compound(6)%component=[ 6 ]
compound(6)%output=.false.

!compound: <T average> of scalar rank
compound(7)%name='<T average>'
compound(7)%units='1'
compound(7)%centring='none'
compound(7)%type='output'
compound(7)%rank='scalar'
compound(7)%region=''
allocate(compound(7)%component(1))
compound(7)%component=[ 7 ]
compound(7)%output=.true.

constants_file = "constants.in" ! filename that was used by setup_equations


! setup someloop which is also used locally within update subroutines
allocate(someloop(6))


if (debug) write(*,'(a/80(1h-))') 'subroutine allocate_meta_arrays'

end subroutine allocate_meta_arrays

!----------------------------------------------------------------------------

subroutine update

! here we update all the dependents and equations

use general_module
integer :: nvar, m, ns, i, j
double precision :: deriv
character(len=1000) :: formatline
logical :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine update'

if (debug) write(*,'(40(1h+)/a)') 'dependent updates'

do nvar = 1, allocatable_size(var_list(var_list_number(centring="all",type="dependent"))%list)
  m = var_list(var_list_number(centring="all",type="dependent"))%list(nvar)

  i = 0
  j = 0

  if (debug) write(*,*) 'updating var dependent: m = ',m,': name = ',trim(var(m)%name),': centring = ',var(m)%centring

! equations
!  <sub_string:dependent>

  if (debug) then
    formatline = '(a,'//trim(indexformat)//',a,i3,a,a)'
    do ns = 1, ubound(var(m)%funk,1)
      write(*,fmt=formatline) '(ns,m) = (',ns,',',m,'): var = ',trim(print_funk(var(m)%funk(ns)))
    end do
  end if

end do

if (debug) write(*,'(a/40(1h-))') 'dependent updates'

!------------------------------------------------------------------

if (debug) write(*,'(40(1h+)/a)') 'equation updates'

do nvar = 1, allocatable_size(var_list(var_list_number(centring="all",type="equation"))%list)
  m = var_list(var_list_number(centring="all",type="equation"))%list(nvar)

  i = 0
  j = 0

  if (debug) write(*,*) 'updating var equation: m = ',m,': name = ',trim(var(m)%name),': centring = ',var(m)%centring

! equations
  if (m == 4) then
! equation <T transport>
! -celldiv(<D>*facegrad(<T>))
! -someloop[1]
    
    do ns = 1, ubound(region(var(m)%region_number)%ij,1)
      i = region(var(m)%region_number)%ij(ns)
      call reset_funk(var(m)%funk(ns))
      call update_someloop(1,i,j)
      
      var(4)%funk(ns)%v = var(4)%funk(ns)%v + (-someloop(1)%v)
      
      deriv = -1
      call add_to_dv(var(4)%funk(ns),deriv,someloop(1))
      
    end do
    
  else if (m == 5) then
! equation <T hole>
! -<D>*facegrad(<T>)-<hole flux>
! -var[2]-var[1]*someloop[2]
    
    do ns = 1, ubound(region(var(m)%region_number)%ij,1)
      j = region(var(m)%region_number)%ij(ns)
      call reset_funk(var(m)%funk(ns))
      call update_someloop(2,i,j)
      
      var(5)%funk(ns)%v = var(5)%funk(ns)%v + (-var(2)%funk(1)%v-var(1)%funk(1)%v*someloop(2)%v)
      
      deriv = -var(1)%funk(1)%v
      call add_to_dv(var(5)%funk(ns),deriv,someloop(2))
      
    end do
    
  else if (m == 6) then
! equation <T walls>
! <T>-1.0d0
! someloop[3]-1.0
    
    do ns = 1, ubound(region(var(m)%region_number)%ij,1)
      j = region(var(m)%region_number)%ij(ns)
      call reset_funk(var(m)%funk(ns))
      call update_someloop(3,i,j)
      
      var(6)%funk(ns)%v = var(6)%funk(ns)%v + (someloop(3)%v-1.0)
      
      deriv = 1
      call add_to_dv(var(6)%funk(ns),deriv,someloop(3))
      
    end do
    
    
  else
    write(*,*) "ERROR: the index ",m," was not found when updating equation"
    stop
  end if


  if (debug) then
    formatline = '(a,'//trim(indexformat)//',a,i3,a,a)'
    do ns = 1, ubound(var(m)%funk,1)
      write(*,fmt=formatline) '(ns,m) = (',ns,',',m,'): var = ',trim(print_funk(var(m)%funk(ns)))
    end do
  end if

end do

if (debug) write(*,'(a/40(1h-))') 'equation updates'

if (debug) write(*,'(a/80(1h-))') 'subroutine update'

end subroutine update

!----------------------------------------------------------------------------

recursive subroutine update_someloop(m,ilast,jlast)

! here we update each of all the possible someloops
! bug wife again!!!

use general_module
integer :: m, ilast, jlast, i, j, ns, region_number, flux_direction
double precision :: deriv
logical :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine update_someloop'

i = ilast
j = jlast

call reset_funk(someloop(m))

if (m == 1) then
! someloop SYSTEM generated sum someloop over face region <nobcelljfaces>
! SYSTEM generated sum someloop over face region <nobcelljfaces>
! var[1]*someloop[6]*divop[i,j]
  
! sum is conducted around region <nobcelljfaces>
! move facesum to domain cell if cell is a boundary cell
  if (cell(i)%type == 2) i = face(cell(i)%jface(1))%icell(1)
  do ns = 1, ubound(cell(i)%jface,1)
    j = cell(i)%jface(ns)
    call update_someloop(6,i,j)
    
    someloop(1)%v = someloop(1)%v + (var(1)%funk(1)%v*someloop(6)%v*divop(i,j))
    
    deriv = var(1)%funk(1)%v*divop(i,j)
    call add_to_dv(someloop(1),deriv,someloop(6))
    
  end do
  
else if (m == 2) then
! someloop SYSTEM generated sum someloop over cell region <facekernel[l=4]>
! SYSTEM generated sum someloop over cell region <facekernel[l=4]>
! var[i,3]*facekernel[j,4,ns]
  
! sum is conducted around region <facekernel[l=4]>
  do ns = 1, ubound(face(j)%kernel(4)%ijk,1)
    i = face(j)%kernel(4)%ijk(ns)
    
    someloop(2)%v = someloop(2)%v + (var(3)%funk(nsvar(3,i))%v*face(j)%kernel(4)%v(ns))
    
    deriv = face(j)%kernel(4)%v(ns)
    call add_to_dv(someloop(2),deriv,var(3)%funk(nsvar(3,i)))
    
  end do
  
else if (m == 3) then
! someloop SYSTEM generated sum someloop over cell region <facekernel[l=0]>
! SYSTEM generated sum someloop over cell region <facekernel[l=0]>
! var[i,3]*facekernel[j,0,ns]
  
! sum is conducted around region <facekernel[l=0]>
  do ns = 1, ubound(face(j)%kernel(0)%ijk,1)
    i = face(j)%kernel(0)%ijk(ns)
    
    someloop(3)%v = someloop(3)%v + (var(3)%funk(nsvar(3,i))%v*face(j)%kernel(0)%v(ns))
    
    deriv = face(j)%kernel(0)%v(ns)
    call add_to_dv(someloop(3),deriv,var(3)%funk(nsvar(3,i)))
    
  end do
  
else if (m == 4) then
! someloop SYSTEM generated sum someloop over cell region <domain>
! SYSTEM generated sum someloop over cell region <domain>
! cellvol[i]*var[i,3]
  
  region_number = region_number_from_name(name='<domain>',centring='cell')
  if (region_number == 0) stop 'ERROR: problem finding cell region <domain> in update_someloops'
  
  do ns = 1, ubound(region(region_number)%ij,1)
    i = region(region_number)%ij(ns)
    
    someloop(4)%v = someloop(4)%v + (cell(i)%vol*var(3)%funk(nsvar(3,i))%v)
  end do
  
else if (m == 5) then
! someloop SYSTEM generated sum someloop over cell region <domain>
! SYSTEM generated sum someloop over cell region <domain>
! cellvol[i]
  
  region_number = region_number_from_name(name='<domain>',centring='cell')
  if (region_number == 0) stop 'ERROR: problem finding cell region <domain> in update_someloops'
  
  do ns = 1, ubound(region(region_number)%ij,1)
    i = region(region_number)%ij(ns)
    
    someloop(5)%v = someloop(5)%v + (cell(i)%vol)
  end do
  
else if (m == 6) then
! someloop SYSTEM generated sum someloop over cell region <facekernel[l=4]>
! SYSTEM generated sum someloop over cell region <facekernel[l=4]>
! var[i,3]*facekernel[j,4,ns]
  
! sum is conducted around region <facekernel[l=4]>
  do ns = 1, ubound(face(j)%kernel(4)%ijk,1)
    i = face(j)%kernel(4)%ijk(ns)
    
    someloop(6)%v = someloop(6)%v + (var(3)%funk(nsvar(3,i))%v*face(j)%kernel(4)%v(ns))
    
    deriv = face(j)%kernel(4)%v(ns)
    call add_to_dv(someloop(6),deriv,var(3)%funk(nsvar(3,i)))
    
  end do
  
  
else
  write(*,*) "ERROR: the index ",m," was not found when updating someloop"
  stop
end if


if (debug) write(*,'(a/80(1h-))') 'subroutine update_someloop'

end subroutine update_someloop

!----------------------------------------------------------------------------

subroutine update_constants

! here we update all the constants that are calculated using equations

use general_module
integer :: nvar, m, ns, i, j
double precision :: deriv
character(len=1000) :: formatline
logical :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine update_constants'

do nvar = 1, allocatable_size(var_list(var_list_number(centring="all",type="constant"))%list)
  m = var_list(var_list_number(centring="all",type="constant"))%list(nvar)

  i = 0
  j = 0

  if (debug) write(*,*) 'updating var constant: m = ',m,': name = ',trim(var(m)%name),': centring = ',var(m)%centring

! equations
!  <sub_string:constant>

  if (debug) then
    formatline = '(a,'//trim(indexformat)//',a,i3,a,a)'
    do ns = 1, ubound(var(m)%funk,1)
      write(*,fmt=formatline) '(ns,m) = (',ns,',',m,'): var = ',trim(print_funk(var(m)%funk(ns)))
    end do
  end if

end do

if (debug) write(*,'(a/80(1h-))') 'subroutine update_constants'

end subroutine update_constants

!----------------------------------------------------------------------------

subroutine update_independents

! here we update all the independent initial values

use general_module
integer :: nvar, m, ns, i, j
double precision :: deriv
character(len=1000) :: formatline
logical :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine update_independents'

do nvar = 1, allocatable_size(var_list(var_list_number(centring="all",type="independent"))%list)
  m = var_list(var_list_number(centring="all",type="independent"))%list(nvar)

  i = 0
  j = 0

  if (debug) write(*,*) 'updating var independent: m = ',m,': name = ',trim(var(m)%name),': centring = ',var(m)%centring

! equations
  if (m == 3) then
! independent <T>
! 1.d0
! 1.0
    
    do ns = 1, ubound(region(var(m)%region_number)%ij,1)
      i = region(var(m)%region_number)%ij(ns)
      call reset_funk(var(m)%funk(ns))
      
      var(3)%funk(ns)%v = var(3)%funk(ns)%v + (1.0)
    end do
    
    
  else
    write(*,*) "ERROR: the index ",m," was not found when updating independent"
    stop
  end if


  if (debug) then
    formatline = '(a,'//trim(indexformat)//',a,i3,a,a)'
    do ns = 1, ubound(var(m)%funk,1)
      write(*,fmt=formatline) '(ns,m) = (',ns,',',m,'): var = ',trim(print_funk(var(m)%funk(ns)))
    end do
  end if

end do

if (debug) write(*,'(a/80(1h-))') 'subroutine update_independents'

end subroutine update_independents

!----------------------------------------------------------------------------

subroutine update_outputs

! here we update all the outputs that are calculated using equations

use general_module
integer :: nvar, m, ns, i, j
double precision :: deriv
character(len=1000) :: formatline
logical :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine update_outputs'

do nvar = 1, allocatable_size(var_list(var_list_number(centring="all",type="output"))%list)
  m = var_list(var_list_number(centring="all",type="output"))%list(nvar)

  i = 0
  j = 0

  if (debug) write(*,*) 'updating var output: m = ',m,': name = ',trim(var(m)%name),': centring = ',var(m)%centring

! equations
  if (m == 7) then
! output <T average>
! cellsum(<T>*<cellvol>,<domain>)/cellsum(<cellvol>,<domain>)
! someloop[4]/someloop[5]
    
    do ns = 1, 1
      call reset_funk(var(m)%funk(ns))
      call update_someloop(4,i,j)
      call update_someloop(5,i,j)
      
      var(7)%funk(ns)%v = var(7)%funk(ns)%v + (someloop(4)%v/someloop(5)%v)
    end do
    
    
  else
    write(*,*) "ERROR: the index ",m," was not found when updating output"
    stop
  end if


  if (debug) then
    formatline = '(a,'//trim(indexformat)//',a,i3,a,a)'
    do ns = 1, ubound(var(m)%funk,1)
      write(*,fmt=formatline) '(ns,m) = (',ns,',',m,'): var = ',trim(print_funk(var(m)%funk(ns)))
    end do
  end if

end do

if (debug) write(*,'(a/80(1h-))') 'subroutine update_outputs'

end subroutine update_outputs

!----------------------------------------------------------------------------

subroutine variable_checker(error)

! here we check on the magnitudes of any variables against specified options

use general_module
integer :: ns, error
logical :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine variable_checker'

!<sub_string:variable_checker>

if (debug) write(*,'(a/80(1h-))') 'subroutine variable_checker'

end subroutine variable_checker

!----------------------------------------------------------------------------

end module equations_module

!----------------------------------------------------------------------------
